import zoneinfo

from flask import current_app
from flask_wtf import FlaskForm
from labbase2.forms import render
from labbase2.forms.validators import (
    AllASCII,
    ContainsLower,
    ContainsNotSpace,
    ContainsNumber,
    ContainsSpecial,
    ContainsUpper,
)
from wtforms.fields import PasswordField, SelectField, StringField, SubmitField
from wtforms.validators import DataRequired, Email, EqualTo, Length

__all__ = ["RegisterForm"]


class RegisterForm(FlaskForm):
    """A registration form for new users.

    Attributes
    ----------
    username : StringField
    email : StringField
    roles : SelectField
    password : PasswordField
    password2 : PasswordField
    submit : SubmitField

    Notes
    -----
    The registration form is thought to be exposed only to an administrator
    of the site since the users should not be allowed to choose their own roles.
    """

    first_name = StringField(
        label="First name",
        validators=[DataRequired(), Length(max=64)],
        render_kw=render.custom_field | {"placeholder": "First name"},
        description="You given name. You may use initials for middle names.",
    )
    last_name = StringField(
        label="Last name",
        validators=[DataRequired(), Length(max=64)],
        render_kw=render.custom_field | {"placeholder": "Last name"},
    )
    email = StringField(
        label="E-Mail Address",
        validators=[DataRequired(), Email(), Length(max=128)],
        render_kw=render.custom_field | {"placeholder": "Email Address"},
        description="The university email address.",
    )
    timezone = SelectField(
        "Timezone",
        choices=[(tz, tz) for tz in sorted(zoneinfo.available_timezones())],
        default=lambda: current_app.config["DEFAULT_TIMEZONE"],
        validators=[DataRequired()],
        render_kw=render.select_field,
        description="""
        Select the timezone in which dates and times shall be displayed for you.
        """,
    )
    password = PasswordField(
        "Password",
        validators=[
            DataRequired(),
            Length(min=12),
            ContainsLower(),
            ContainsUpper(),
            ContainsNumber(),
            ContainsSpecial(),
            ContainsNotSpace(),
            AllASCII(),
        ],
        render_kw=render.custom_field | {"placeholder": "Password"},
        description="""
        Minimum 12 characters. Contains lower- and uppercase characters. Contains at 
        least 1 number. Contains a special character. Does not contain spaces. Only 
        ASCII characters.
        """,
    )
    password2 = PasswordField(
        "Repeat Password",
        validators=[DataRequired(), EqualTo("password")],
        render_kw=render.custom_field | {"placeholder": "Repeat Password"},
    )
    submit = SubmitField("Submit", render_kw=render.submit_field)



from labbase2.models import db
from labbase2.models.consumable import Consumable
from labbase2.models.mixins.filter import Filter
from sqlalchemy import asc, desc, func
from sqlalchemy.orm import subqueryload

__all__ = ["Chemical", "StockSolution"]


class Chemical(Consumable):
    """A class to represent a chemical compound.

    Currently, a Chemical has only very few attributes, i.e., two: The CAS
    number and the PubChem CID. The CID is used to retrieve further
    informations from the web. The CAS number is sued for 'official' reference.

    Attributes
    ----------
    id : int
        The internal ID of this chemical. This is unique among all entities
        in the database.
    responsible_id : int
        ID of the user who is responsible for this chemical.
    cas_number : str
        The CAS registry number, a mandatory number for all chemical compounds.
    pubchem_cid : int
        The PubChem reference number. This will be used to retrieve
        additional informations about this compound from PubChem.

    Notes
    -----
    Currently, the Chemical class uses the PubChem CID to retrieve
    information like the molecular weight from the PubChem REST API. This
    causes problems at the moment. These problems might be caused of people
    requesting too much information in too little time followed by banning the IP
    by PubChem. Therefor, it is likely that the implementation will change in
    the future such that people have to add specific information about the
    chemical themselves.
    """

    __tablename__: str = "chemical"

    id = db.Column(
        db.Integer,
        db.ForeignKey("consumable.id"),
        primary_key=True,
        info={"importable": False},
    )
    molecular_weight = db.Column(db.Float, info={"importable": True})

    # One-to-many relationships.
    stocks = db.relationship(
        "StockSolution",
        backref="chemical",
        lazy=True,
        order_by="StockSolution.date_emptied, StockSolution.date_created.desc()",
    )

    __mapper_args__ = {"polymorphic_identity": "chemical"}

    @classmethod
    def _options(cls) -> tuple:
        return subqueryload(cls.stocks), subqueryload(cls.batches)


class StockSolution(db.Model, Filter):
    """

    Attributes
    ----------
    id : int
        An id that identifies the stock solution.
    label : str
        The label, that is the token, of the stock solution. Can be something
        like 'Tris/HCl pH 9.5'.
    details : str
        A mored detailed description of the stock solution.

    Notes
    -----
    Stocks are used to model stock solutions. Each stock solution can contain
    several chemicals. Therefor, this is modeled as a many-to-many
    relationship.
    """

    id = db.Column(db.Integer, primary_key=True)
    chemical_id = db.Column(db.Integer, db.ForeignKey("chemical.id"), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    solvent = db.Column(db.String(64), nullable=False)
    date_created = db.Column(db.Date, nullable=False, default=func.today())
    date_emptied = db.Column(db.Date, nullable=True)
    concentration = db.Column(db.String(32), nullable=False)
    storage_place = db.Column(db.String(64), nullable=False)
    details = db.Column(db.String(2048))

    @classmethod
    def _filters(cls, **fields) -> list:
        if lbl := fields.pop("label", None):
            f = super()._filters(**fields) + [Chemical.label.ilike(f"%{lbl}%")]
        else:
            f = super()._filters(**fields)

        return f

    @classmethod
    def _order_by(cls, order_by: str, ascending: bool) -> tuple:
        fnc = asc if ascending else desc
        if order_by == "label":
            field = Chemical.label
        else:
            field = getattr(cls, order_by.strip())

        return (fnc(field),)

    @classmethod
    def _entities(cls) -> tuple:
        return cls, Chemical

    @classmethod
    def _joins(cls) -> tuple:
        return ((Chemical, cls.chemical_id == Chemical.id),)



from datetime import date

from labbase2.models import db
from labbase2.models.base_entity import BaseEntity
from labbase2.models.mixins.export import Export
from labbase2.models.mixins.filter import Filter
from sqlalchemy import asc, desc, not_
from sqlalchemy.orm import column_property

__all__ = ["Batch", "Consumable"]


class Batch(db.Model, Filter, Export):
    """A batch is an order of a consumable.

    Attributes
    ----------
    id : int
        The internal database identifier of this batch.
    consumable_id : int
        The identifier of the consumable to which this batch belongs.
    supplier : str
        The company (or whatever) where this batch was ordered (again, the same
        chemical might be ordered at different companies due to discounts,
        for instance).
    article_number : str
        The article number of the ordered consumable. This serves as a unambiguous
        identifier of the ordered chemical. This is important for different quality
        grades of the same chemical, for instance.
    amount : str
        The amoint that was ordered. For some companies this is already indicated by
        the article number but for some is not. The information is straightforward
        for chemicals. For enzymes this might be the number of reactions or the total
        activity in U. For antibodies the total volume along with the concentration
        would suffice.
    date_ordered : date
        The date at which this batch was ordered.
    date_opened : date
        The date at which this batch was opened.
    date_expiration : date
        The expiration date that was set by the supplier.
    date_emptied : date
        The date at which this batch was emptied.
    price : float
        The price in euros.
    storage_place : str
        The location where this batch can be found. This should be clear and
        unambiguous.
    lot : str
        The lot number of this batch.
    in_use : bool
        A flag indicating if this batch is currently in use. At best, only ever one
        batch for each consumable is in use at a time.
    is_open : bool
        A flag indicating if this batch was opened already.
    is_empty : bool
        A flag indicating if this batch is empty.

    Notes
    -----
    While a consumable is just a description a batch is the actual thing available in
    the lab. For instance, an entry for the chemical 'SDS' only gives general
    informations about the chemical and that it is or was available in the lab at
    some point in time. However, a batch is the actual order of that consumable
    along with additional informations that might be of interest like article number
    and source (the same chemical might be bought from different manufacturers).
    """

    id = db.Column(db.Integer, primary_key=True)
    consumable_id = db.Column(
        db.Integer, db.ForeignKey("consumable.id"), nullable=False
    )
    supplier = db.Column(db.String(64), nullable=False)
    article_number = db.Column(db.String(32), nullable=False)
    amount = db.Column(db.String(32))
    date_ordered = db.Column(db.Date)
    date_opened = db.Column(db.Date)
    date_expiration = db.Column(db.Date)
    date_emptied = db.Column(db.Date)
    price = db.Column(db.Float())
    storage_place = db.Column(db.String(64), nullable=False)
    lot = db.Column(db.String(64), nullable=False)
    in_use = db.Column(db.Boolean, nullable=False, default=False)

    is_open = column_property(date_opened.isnot(None).label("is_open"), deferred=True)
    is_empty = column_property(
        date_emptied.isnot(None).label("is_empty"), deferred=True
    )

    @classmethod
    def _filters(cls, **fields) -> list:
        filters = []

        if type_ := fields.pop("consumable_type", None):
            filters.append(Consumable.entity_type.is_(type_))
        if label := fields.pop("label", None):
            filters.append(Consumable.label.ilike(f"%{label}%"))

        if (empty := fields.pop("empty", "all")) == "empty":
            filters.append(cls.is_empty)
        elif empty != "all":
            filters.append(not_(cls.is_empty))

        if (in_use := fields.pop("in_use", "all")) == "in_use":
            filters.append(cls.in_use)
        elif in_use != "all":
            filters.append(not_(cls.in_use))

        return super()._filters(**fields) + filters

    @classmethod
    def _order_by(cls, order_by: str, ascending: bool) -> tuple:
        match order_by.strip():
            case "id":
                field = cls.id
            case "label":
                field = Consumable.label
            case "consumable_type":
                field = Consumable.entity_type
            case "supplier":
                field = cls.supplier
            case "date_ordered":
                field = cls.date_ordered
            case _:
                raise ValueError("Unknown order field!")

        fnc = asc if ascending else desc

        return (fnc(field),)

    @classmethod
    def _entities(cls) -> tuple:
        return cls, Consumable.entity_type, Consumable.label

    @classmethod
    def _joins(cls) -> tuple:
        return (Consumable,)


class Consumable(BaseEntity, Export):
    """A consumable is a general class for all kind of stuff in the lab that can be
    used up.

    The thing about consumables is that they can have batches attached to them.
    Please note that a consumable is just a description of the respective enzyme,
    chemical, etc. It does give no information about availability in the lab. For this
    information one has to consult the batches.

    Attributes
    ----------
    id : int
        The internal identifier of the consumable.
    storage_info : str
        A short description how this consumable should be stored.
    batches : list[Batch]
        A list of all batches of this consumable that were ordered.
    """

    id = db.Column(
        db.Integer,
        db.ForeignKey("base_entity.id"),
        primary_key=True,
        info={"importable": False},
    )
    storage_info = db.Column(db.String(64), info={"importable": True})

    # One-to-many relationships.
    batches = db.relationship(
        "Batch",
        backref="consumable",
        lazy=True,
        order_by="Batch.date_emptied, Batch.in_use.desc(), Batch.date_ordered",
    )

    # Proper setup for joined table inheritance.
    __mapper_args__ = {"polymorphic_identity": "consumable"}

    def to_dict(self) -> dict:
        return super().to_dict() | {"batches": [b.to_dict() for b in self.batches]}

    @property
    def location(self):
        if self.batches:
            return self.batches[0].storage_place
        else:
            return None

    @classmethod
    def _subquery(cls):
        return (
            db.session.query(Batch.consumable_id, Batch.storage_place)
            .filter(Batch.in_use)
            .subquery(name="batch")
        )



from flask_wtf import FlaskForm
from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import EditEntityForm, FilterForm
from labbase2.models import Chemical, User
from sqlalchemy import func
from wtforms.fields import (
    DecimalField,
    Field,
    IntegerField,
    SelectField,
    SelectMultipleField,
    StringField,
)
from wtforms.validators import Length, NumberRange, Optional
from wtforms.widgets import CheckboxInput, ListWidget

__all__: list = ["FilterChemical", "EditChemical"]


class FilterChemical(FilterForm):
    """Form for searching chemicals in the database.

    Attributes
    ----------
    cas : StringField
        The CAS number or part of it.
    pubchem_cid : StringField
        The PubChem CID. Please do not confuse this with the PubChem SID that
        also exists.
    order_by : SelectField
        The attribute by which the results shall be ordered. This overrides
        the order_field field of the parent SearchBaseForm class.
    """

    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Label"},
        description="The label, i.e. the name of the entity.",
    )
    owner_id = SelectField(
        "Responsible",
        choices=[(0, "All")],
        default=0,
        coerce=int,
        render_kw=render.select_field,
    )
    order_by = SelectField(
        label="Order by",
        choices=[("label", "Label"), ("id", "ID"), ("order_date", "Order date")],
        default="label",
        render_kw=render.select_field,
        description="The column by which the results shall be ordered.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        user = (
            User.query.join(Chemical)
            .with_entities(User.id, User.username)
            .group_by(User.username)
            .having(func.count(User.id) > 0)
            .order_by(User.username)
        )
        self.owner_id.choices += user

    def fields(self) -> list:
        return [
            self.id,
            self.label,
            self.owner_id,
            self.order_by,
            self.ascending,
        ]


class EditChemical(EditEntityForm):
    """Form to edit chemicals.

    Attributes
    ----------
    cas_number : StringField
        The CAS registry number of the chemical.
    pubchem_cid : IntegerField
        The PubChem CID of the chemical. This should not be confused with the
        PubChem SID.
    """

    molecular_weight = DecimalField(
        "Molecular Weight",
        validators=[Optional(), NumberRange(min=0)],
        render_kw=render.custom_field | {"placeholder": "Molecular weight"},
    )
    owner_id = SelectField(
        "User",
        choices=[(0, "All")],
        coerce=int,
        render_kw=render.select_field,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        user = (
            User.query.with_entities(User.id, User.username)
            .order_by(User.username)
            .all()
        )
        self.owner_id.choices += user



from flask import Blueprint, flash, render_template, request
from flask_login import current_user, login_required
from labbase2.models import GlycerolStock, Plasmid, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.files.forms import UploadFile
from labbase2.views.plasmids.forms import EditPlasmid

from .forms import EditBacterium, FilterBacteria

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "bacteria", __name__, url_prefix="/glycerol-stocks", template_folder="templates"
)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterBacteria(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = GlycerolStock.filter_(**data)
    except Exception as error:
        flash(str(error), "danger")
        entities = GlycerolStock.filter_(order_by="label")

    return render_template(
        "bacteria/main.html",
        filter_form=form,
        add_form=EditBacterium(formdata=None),
        entities=entities.paginate(page=page, per_page=100),
        total=GlycerolStock.query.count(),
        title="Glycerol stocks",
    )


@bp.route("/<int:plasmid_id>", methods=["POST"])
@login_required
@permission_required("Add glycerol stocks")
def add(plasmid_id: int):
    form = EditBacterium()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (plasmid := Plasmid.query.get(plasmid_id)) is None:
        return f"No plasmid with ID {plasmid_id}!"

    stock = GlycerolStock(plasmid_id=plasmid_id, owner_id=current_user.id)
    form.populate_obj(stock)

    try:
        db.session.add(stock)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully added glycerol stock with '{plasmid.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add glycerol stocks")
def edit(id_: int):
    form = EditBacterium()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if not (plasmid := GlycerolStock.query.get(id_)):
        return Message.ERROR(f"No glycerol stock with ID {id_}!")

    if plasmid.owner_id != current_user.id:
        return Message.ERROR("Glycerol stocks can only be edited by owner!")

    form.populate_obj(plasmid)

    try:
        db.session.commit()
    except Exception as err:
        db.session.rollback()
        return Message.ERROR(str(err))

    return Message.SUCCESS(f"Successfully edited glycerol stock!")


@bp.route("/<int:id_>/<string:format_>", methods=["GET"])
@login_required
def details(id_: int, format_: str):
    if (stock := GlycerolStock.query.get(id_)) is None:
        return Message.ERROR(f"No glycerol stock with ID {id_}!")

    edit_form = EditBacterium(None, obj=stock)

    match format_:
        case "long":
            template = "bacteria/details.html"
        case "tab":
            template = "bacteria/details-tab.html"
        case _:
            return Message.ERROR(f"Invalid format '{format_}'!")

    return render_template(
        template,
        stock=stock,
        form=edit_form,
        file_form=UploadFile,
        plasmid_form=EditPlasmid(None, obj=stock.plasmid),
    )


@bp.route("/glycerol-stock/export/<string:format_>/", methods=["GET"])
@login_required
@permission_required("Export content")
def export(format_: str):
    data = FilterBacteria(request.args).data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = GlycerolStock.filter_(**data)
    except Exception as error:
        return Message.ERROR(error)

    match format_:
        case "csv":
            return GlycerolStock.to_csv(entities)
        case "json":
            return GlycerolStock.to_json(entities)
        case _:
            return Message.ERROR(f"Unsupported format '{format_}'!")



import io
from datetime import date

from Bio import SeqIO
from Bio.Restriction import CommOnly
from flask import send_file

__all__ = ["Sequence"]


class Sequence:
    """
    A mixin to export sequence entities (primers/plasmids) to FASTA files.
    """

    def __len__(self):
        return len(self.seqrecord.seq)

    @property
    def seqrecord(self):
        raise NotImplementedError

    def restriction_sites(self) -> dict:
        if record := self.seqrecord:
            sites = CommOnly.search(record.seq, linear=False)
            sites = {k: v for k, v in sites.items() if len(v) == 1}
            sites = [k for k in sites]
            sites.sort(key=lambda x: x.__name__)

            return sites
        else:
            return {}

    def formatted_restriction_sites(self) -> str:
        """Return restriction sites with HTML markup.

        Returns
        -------
        str
            A single string that can be inserted into a website as it is. It
            highlights the number of cutting sites in bold.
        """

        sites = [(e, n) for e, n in self.restriction_sites().items()]
        sites.sort(key=lambda x: (x[1], x[0]))

        return ", ".join(["{} <b>({})</b>".format(e, n) for e, n in sites])

    @classmethod
    def to_fasta(cls, instances):
        with io.StringIO() as proxy:
            SeqIO.write([i.seqrecord for i in instances], proxy, "fasta")
            mem = io.BytesIO(proxy.getvalue().encode("utf-8"))

        filename = cls.__name__ + "_" + date.today().isoformat() + ".fasta"

        return send_file(
            mem, as_attachment=True, download_name=filename, mimetype="text/fasta"
        )



from labbase2.models import db
from labbase2.models.mixins.export import Export
from labbase2.models.mixins.importer import Importer
from sqlalchemy import func

__all__ = ["Comment"]


class Comment(db.Model, Importer, Export):
    """A comment to an entity.

    Attributes
    ----------
    id : int
        The identifier of this comment.
    entity_id : int
        The identifier of the entity about which this comment is.
    user_id : str
        The id of the person that has written the comment.
    timestamp_created : DateTime
        The time of the comment. This is automatically set by the database.
    timestamp_edited : DateTime
        The time this comment was last edited. Might be `None` if the comment was
        never modified.
    subject : str
        A short string describing the subject of the comment.
    text : str
        The message of the comment.
    """

    __tablename__: str = "comment"

    id = db.Column(db.Integer, primary_key=True, info={"importable": False})
    entity_id = db.Column(
        db.Integer,
        db.ForeignKey("base_entity.id"),
        nullable=False,
        info={"importable": True},
    )
    user_id = db.Column(
        db.Integer, db.ForeignKey("user.id"), nullable=False, info={"importable": True}
    )
    timestamp_created = db.Column(
        db.DateTime,
        nullable=False,
        server_default=func.now(),
        info={"importable": True},
    )
    timestamp_edited = db.Column(
        db.DateTime(timezone=True),
        nullable=True,
        onupdate=func.now(),
        info={"importable": True},
    )
    subject = db.Column(db.String(128), nullable=False, info={"importable": True})
    text = db.Column(db.String(2048), nullable=False, info={"importable": True})

    __table_args__ = {"extend_existing": True}



from functools import wraps
from typing import Callable

from flask import flash, redirect, url_for
from flask_login import current_user
from labbase2.models import db, Permission

__all__ = ["permission_required"]


def permission_required(*allowed) -> Callable:
    """Check whether the current user has sufficient role to access a resource.

    This is a very simple decorator op to add user role system to the
    website.

    Parameters
    ----------
    *allowed : *str
        A list of permissions. If the user has any of these permissions, he will be
        granted access to the view.

    Returns
    -------
    function
        The decorate route function.
    """

    def decorator(func: Callable):

        @wraps(func)
        def decorated_view(*args, **kwargs):

            verified = []

            for permission in allowed:
                permission_db = db.session.get(Permission, permission)
                if permission_db is None:
                    flash(
                        f"Permission '{permission}' not found." f"Please inform the developer!",
                        "warning",
                    )
                else:
                    verified.append(permission_db)

            # Check if allowed and actual roles have non-empty intersection.
            if current_user.is_admin:
                return func(*args, **kwargs)

            if not verified:
                flash("No valid permissions defined for this site!", "danger")
                return redirect(url_for("base.index"))

            for permission in current_user.permissions:
                if permission in verified:
                    return func(*args, **kwargs)

            flash("No permission to enter this site!", "warning")
            return redirect(url_for("base.index"))

        return decorated_view

    return decorator



from flask import Blueprint
from flask import current_app as app
from flask import flash, redirect, render_template, request
from flask_login import current_user, login_required
from labbase2.models import BaseFile, Plasmid, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.comments.forms import EditComment
from labbase2.views.files.forms import UploadFile
from labbase2.views.files.routes import upload_file
from labbase2.views.requests.forms import EditRequest

from . import bacteria, preparations
from .bacteria.forms import EditBacterium
from .forms import EditPlasmid, FilterPlasmids
from .preparations.forms import EditPreparation

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint("plasmids", __name__, url_prefix="/plasmid", template_folder="templates")

bp.register_blueprint(bacteria.bp)
bp.register_blueprint(preparations.bp)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterPlasmids(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Plasmid.filter_(**data)
    except Exception as error:
        flash(str(error), "danger")
        entities = Plasmid.filter_(order_by="label")

    return render_template(
        "plasmids/main.html",
        filter_form=form,
        import_file_form=UploadFile(),
        add_form=EditPlasmid(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=Plasmid.query.count(),
        title="Plasmids",
    )


@bp.route("/", methods=["POST"])
@login_required
@permission_required("Add plasmid")
def add():
    form = EditPlasmid()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    plasmid = Plasmid(owner_id=current_user.id)
    form.populate_obj(plasmid)

    try:
        db.session.add(plasmid)
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully added plasmid '{plasmid.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add plasmid")
def edit(id_: int):
    form = EditPlasmid()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (plasmid := Plasmid.query.get(id_)) is None:
        return Message.ERROR(f"No plasmid with ID {id_}!")

    if plasmid.owner_id != current_user.id:
        return Message.ERROR("Plasmids can only be edited by owner!")

    form.populate_obj(plasmid)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully edited plasmid!")


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add plasmid")
def delete(id_: int):
    if (plasmid := Plasmid.query.get(id_)) is None:
        return Message.ERROR(f"No plasmid with ID {id_}!")

    if plasmid.owner_id != current_user.id:
        return Message.ERROR("Only owner can delete plasmid!")

    try:
        db.session.delete(plasmid)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully deleted plasmid '{plasmid.label}'!")


@bp.route("/<int:id_>", methods=["GET"])
@login_required
def details(id_: int):
    if (plasmid := Plasmid.query.get(id_)) is None:
        return Message.ERROR(f"No plasmid with ID {id_}!")

    return render_template(
        "plasmids/details.html",
        plasmid=plasmid,
        form=EditPlasmid(None, obj=plasmid),
        file_form=UploadFile,
        comment_form=EditComment,
        request_form=EditRequest,
        preparation_form=EditPreparation,
        bacteria_form=EditBacterium,
    )


@bp.route("/<int:id_>/upload/<string:type_>/", methods=["POST"])
@login_required
@permission_required("Upload files")
def upload_plasmid_file(id_: int, type_: str):
    form = UploadFile()

    if (plasmid := Plasmid.query.get(id_)) is None:
        flash(f"No plasmid with ID {id_}!", "danger")
        return redirect(request.referrer)

    if plasmid.owner_id != current_user.id:
        flash(f"Only plasmid owner can upload files!", "danger")
        return redirect(request.referrer)

    if not form.validate():
        flash("No file was uploaded!", "danger")
        return redirect(request.referrer)

    file = upload_file(form, BaseFile)

    match type_:
        case "file":
            if plasmid.file:
                db.session.delete(plasmid.file)
            plasmid.file = file
            flash("Successfully uploaded plasmid file!", "success")
        case "map":
            if plasmid.map:
                db.session.delete(plasmid.map)
            plasmid.map = file
            flash("Successfully uploaded plasmid map!", "success")
        case _:
            flash(f"Unknown type {type_}!", "danger")
            db.session.delete(file)

    db.session.commit()

    return redirect(request.referrer)


@bp.route("/export/<string:format_>/", methods=["GET"])
@login_required
@permission_required("Export content")
def export(format_: str):
    data = FilterPlasmids(request.args).data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Plasmid.filter_(**data)
    except Exception as error:
        return Message.ERROR(error)

    match format_:
        case "csv":
            return Plasmid.export_to_csv(entities)
        case "json":
            return Plasmid.export_to_json(entities)
        case "pdf":
            return Plasmid.to_pdf(entities)
        # case "fasta":
        #     return Plasmid.to_fasta(entities)
        case "zip":
            return Plasmid.to_zip(entities)
        case _:
            return Message.ERROR(f"Unsupported format '{format_}'!")



__all__ = ["DefaultConfig"]


class DefaultConfig(object):
    SECRET_KEY: str = "verysecretkey"

    # Database.
    SQLALCHEMY_DATABASE_URI: str = "sqlite:///labbase2.db"
    SQLALCHEMY_TRACK_MODIFICATIONS: bool = False

    # Display.
    DELETABLE_HOURS: int = 72
    PER_PAGE: int = 100

    # User.
    USER: tuple = "Max", "Mustermann", "test@test.de"
    DEFAULT_TIMEZONE: str = "Europe/Berlin"
    UPLOAD_FOLDER: str = "upload"
    RESET_EXPIRES: int = 1
    PERMISSIONS: list[tuple[str, str]] = [
        (
            "Write comment",
            """Allows a user to author own comments. Applies to all entity types that 
            support comments. Suggested level: Intern.""",
        ),
        (
            "Upload files",
            """Allows a user to upload files. Applies to all entity types that support 
            files. Suggested level: Master student.""",
        ),
        (
            "Add dilutions",
            """Allows a user to add dilutions to antibodies. Suggested level: Bachelor 
            student.""",
        ),
        (
            "Add preparations",
            """Allows a user to add plasmid preparations. Suggested level: Intern.""",
        ),
        (
            "Add glycerol stocks",
            """Allows a user to create glycerol stocks. Suggested level: PhD 
            student.""",
        ),
        (
            "Add consumable batches",
            """Allows a user to add batches of consumables. Suggested level: 
            Technician.""",
        ),
        (
            "Add antibodies",
            """Allows a user to add antibodies. Suggested level: Master/PhD student.""",
        ),
        (
            "Add chemicals",
            """Allows a user to add chemicals. Suggested level: PhD student.""",
        ),
        (
            "Add stock solutions",
            """Allows a user to create glycerol stocks. Suggested level: PhD 
            student.""",
        ),
        (
            "Add Fly Stock",
            """Allows a user to create fly stocks. Suggested level: PhD""",
        ),
        (
            "Add plasmid",
            """Allows a user to add cloned plasmids. Suggested level: Bachelor 
            student.""",
        ),
        (
            "Add oligonucleotide",
            """Allows a user to add primer/oligonucleotides. Suggested level: Bachelor 
            student.""",
        ),
        (
            "Manage users",
            """Allows a user to manage other users. This includes registering new users 
            as well as changing a users status and permissions.""",
        ),
        (
            "Export content",
            """Allows a user to export content. Suggested level: PhD Student.""",
        ),
        (
            "Add requests",
            """Allows a user to add requests for any ressource. Suggested level: PI.""",
        ),
    ]

    # Data.
    RESISTANCES: list[str] = [
        "Ampicillin",
        "Ampicillin and Kanamycin",
        "Apramycin",
        "Chloramphenicol",
        "Chloramphenicol and Ampicillin",
        "Gentamicin",
        "Kanamycin",
        "Spectinomycin",
        "Streptomycin",
        "Tetracyclin",
    ]
    STRAINS: list[str] = [
        "ccdB Survival 2 T1",
        "DB3.1",
        "DH10B",
        "DH5alpha",
        "HB101",
        "JM109",
        "JM110",
        "MC1061",
        "MG1655",
        "NEB Stable",
        "Pir1",
        "Stbl2",
        "Stbl3",
        "Stbl4",
        "TOP10",
        "XL1 Blue",
        "XL10 Gold",
    ]



from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm
from wtforms.fields import FileField, StringField, TextAreaField
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["UploadFile"]


class UploadFile(BaseForm):
    """Form to upload files.

    Attributes
    ----------

    """

    file = FileField(
        "Select File", validators=[DataRequired()], render_kw=render.file_field
    )
    filename = StringField(
        "Filename",
        validators=[Optional(), Length(max=64)],
        render_kw=render.custom_field | {"placeholder": "(Optional)"},
        description="Choose an optional filename.",
    )
    note = TextAreaField(
        "Note",
        validators=[Optional(), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Note", "rows": 8},
    )



from flask import current_app
from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm, FilterForm
from labbase2.models.plasmid import GlycerolStock
from wtforms.fields import DateField, SelectField, StringField
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["FilterBacteria", "EditBacterium"]


class FilterBacteria(FilterForm):
    strain = SelectField(
        label="Strain",
        validators=[Optional()],
        choices=[("all", "All")],
        render_kw=render.select_field | {"placeholder": "DH10B"},
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        strains = (
            GlycerolStock.query.with_entities(
                GlycerolStock.strain, GlycerolStock.strain
            )
            .order_by(GlycerolStock.strain)
            .all()
        )
        self.strain.choices += strains

    def fields(self):
        return [self.id, self.strain, self.order_by, self.ascending]


class EditBacterium(BaseForm):
    """A form for adding or editing a bacterial stock.

    Attributes
    ----------
    strain : SelectField
        The bacterial strain that was transformed.
    transformation_date : DateField
        The date the bacterial stock was generated. This is the day the
        bacteria got transformed.
    disposal_date : DateField
        The date the bacterial stock was trashed.
    storage_place : StringField
        The storage location (exact position in the -80°C freezer) of the
        bacterial stock.
    """

    strain = SelectField(
        label="Strain",
        validators=[DataRequired()],
        choices=[],
        default="DH10B",
        render_kw=render.select_field,
    )
    transformation_date = DateField(
        label="Transformation date",
        validators=[DataRequired()],
        render_kw=render.custom_field | {"type": "date"},
    )
    storage_place = StringField(
        label="Storage place",
        validators=[DataRequired(), Length(max=128)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Storage place"},
    )
    disposal_date = DateField(
        label="Disposal date",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.strain.choices = [
            (strain, strain) for strain in current_app.config["STRAINS"]
        ]



from flask_wtf import FlaskForm
from labbase2.forms import render
from labbase2.forms.validators import (
    AllASCII,
    ContainsLower,
    ContainsNotSpace,
    ContainsNumber,
    ContainsSpecial,
    ContainsUpper,
)
from wtforms.fields import PasswordField, SubmitField
from wtforms.validators import DataRequired, EqualTo, Length, Optional

__all__ = ["ChangePassword"]


class ChangePassword(FlaskForm):

    new_password = PasswordField(
        "New password",
        validators=[
            DataRequired(),
            Length(min=12),
            ContainsLower(),
            ContainsUpper(),
            ContainsNumber(),
            ContainsSpecial(),
            ContainsNotSpace(),
            AllASCII(),
        ],
        render_kw=render.custom_field | {"placeholder": "new password"},
        description="""
        Minimum 12 characters. Contains lower- and uppercase characters. Contains at 
        least 1 number. Contains a special character. Does not contain spaces. Only 
        ASCII characters.
        """,
    )
    new_password2 = PasswordField(
        "Repeat new password",
        validators=[DataRequired(), EqualTo("new_password")],
        render_kw=render.custom_field | {"placeholder": "repeat new password"},
    )
    old_password = PasswordField(
        "Old password",
        validators=[Optional()],
        render_kw=render.custom_field | {"placeholder": "old password"},
    )
    submit = SubmitField("Submit", render_kw=render.submit_field)



from labbase2.models import db
from labbase2.models.mixins.importer import Importer
from sqlalchemy import func

__all__ = ["Request"]


class Request(db.Model, Importer):
    """A table of requests for a specific entity. This is most useful for fly
    stocks, self-made antibodies, and plasmids.

    Attributes
    ----------
    id : int
        The internal id of the request.
    entity_id : int
        The id of the entity that was requested.
    requested_by : str
        Name of the person that has asked for the entity.
    timestamp : Date
        Date of the request. This defaults to the current date.
    timestamp_sent : date
        The date at which the entity was sent to the requester.

    Notes
    -----
    The whole point about the BaseEntity class is to have a single implementation for
    comments and files that are interesting for all entities. Requests are special in
    this sense as it does not make sense for all kind of child classes. Commercially
    available antibodies or chemicals, for instance, will never be requested.
    However, the BaseEntity class is still the least common denominator to all
    classes, for which requests are of interest.
    """

    id = db.Column(db.Integer, primary_key=True, info={"importable": False})
    entity_id = db.Column(
        db.Integer,
        db.ForeignKey("base_entity.id"),
        nullable=False,
        info={"importable": True},
    )
    requested_by = db.Column(db.String(128), nullable=False, info={"importable": True})
    timestamp = db.Column(
        db.Date, server_default=func.today(), info={"importable": True}
    )
    timestamp_sent = db.Column(db.Date, info={"importable": True})
    note = db.Column(db.String(2048), nullable=True, info={"importable": True})

    __table_args__ = {"extend_existing": True}

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "entity_id": self.entity_id,
            "requested_by": self.requested_by,
            "timestamp": self.timestamp.isoformat(),
            "sent": self.sent.isoformat() if self.sent else None,
        }



from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import EditEntityForm, FilterForm
from labbase2.models import User
from wtforms.fields import DateField, SelectField, StringField, TextAreaField
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["FilterPlasmids", "EditPlasmid"]


class FilterPlasmids(FilterForm):
    """Form for searching plasmids in the database.

    Attributes
    ----------
    insert : StringField
        The insert of the plasmid. Usually, plasmids can be identified by the
        vector (backbone) and the insert.
    vector : StringField
        The vector of the plasmid. Usually, plasmids can be identified by the
        vector (backbone) and the insert.
    description : StringField
        Some plasmids have a description informing about what the plasmid
        is/was used for. This can be searched. However, search queries should be
        restricted to single tags because the search is exact.
    order_by : SelectField
        The attribute by which the results shall be ordered. This overrides
        the order_field field of the parent SearchBaseForm class.
    """

    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Label"},
        description="The label, i.e. the name of the entity.",
    )
    insert = StringField(
        label="Insert",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "aka"},
        description="""
        The construct that was inserted into the vector. The search is case
        insensitive.
        """,
    )
    vector = StringField(
        label="Vector",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "pUASt-attB"},
        description="The vector, i.e., the backbone of the plasmid.",
    )
    owner_id = SelectField(
        label="Owner",
        validators=[Optional()],
        choices=[(0, "All")],
        default=0,
        coerce=int,
        render_kw=render.select_field,
        description="The owner of the primer.",
    )
    description = StringField(
        label="Description",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "anne uv flyfos"},
        description="""
        A whitespace separated list of tags in the description of the plasmid. For 
        instance, "anne uv flyfos" finds all plasmids with a description that 
        contains the words "anne", "uv", <strong>AND</strong> "flyfos".
        """,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.order_by.choices += [
            ("label", "Label"),
            ("vector", "Vector"),
            ("cloning_date", "Cloning date"),
        ]
        users = (
            User.query.with_entities(User.id, User.username)
            .order_by(User.username)
            .all()
        )
        self.owner_id.choices += users

    def fields(self) -> list:
        return [
            self.id,
            self.label,
            self.insert,
            self.vector,
            self.owner_id,
            self.description,
            self.order_by,
            self.ascending,
        ]


class EditPlasmid(EditEntityForm):
    """Form to add or edit a plasmid.

    Attributes
    ----------
    vector : StringField
        Limited to 256 characters.
    insert : StringField
        Limited to 128 characters.
    cloning_date : DateField
    description : TextAreaField
        Limited to 1024 characters.
    reference : StringField
        Limited to 512 characters.
    """

    insert = StringField(
        label="Insert",
        validators=[DataRequired(), Length(max=128)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Insert"},
    )
    vector = StringField(
        label="Vector",
        validators=[Optional(), Length(max=256)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Vector"},
    )
    cloning_date = DateField(
        label="Cloning date",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    description = TextAreaField(
        label="Description",
        validators=[Optional(), Length(max=1024)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Description", "rows": 4},
    )
    reference = StringField(
        label="Reference",
        validators=[Optional(), Length(max=512)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Reference"},
    )



__all__ = ["custom_field", "select_field", "file_field", "submit_field"]


custom_field: dict[str, str] = {"class": "form-control form-control-sm"}
select_field: dict[str, str] = {"class": "form-select form-select-sm", "size": 1}
boolean_field: dict[str, str] = {"class": "form-check-input"}
file_field: dict[str, str] = {"class": "form-control form-control-sm"}
submit_field: dict[str, str] = {"class": "btn btn-sm btn-primary"}



from datetime import date

from labbase2.models import db
from sqlalchemy.types import TypeDecorator


class CustomDate(TypeDecorator):

    impl = db.Date

    def process_bind_param(self, value, dialect):
        if isinstance(value, str):
            try:
                return date.fromisoformat(value)
            except:
                raise ValueError(f"Could not parse date {value}!")

        return value


class SequenceString(TypeDecorator):

    impl = db.String

    def process_bind_param(self, value, dialect):
        if value is None:
            return None

        value = value.upper()

        # Remove any whitespace characters from the sequence.
        return "".join(value.split())

    def copy(self, **kw):
        return SequenceString(self.impl.length)



from flask_wtf import FlaskForm
from labbase2.forms import render
from labbase2.forms.filters import make_lower, strip_input
from wtforms.fields import BooleanField, PasswordField, StringField, SubmitField
from wtforms.validators import DataRequired


class LoginForm(FlaskForm):
    """The form used on the login site for users to log in.

    Attributes
    ----------
    email : StringField
        The email of the user. Please note that the email has to be unique among all
        users. Therefore, it can be used as an identifier. The username does not have
        to be unique. Accordingly, it is not suitable for logging in.
    password : PasswordField
        The current password of the user.
    remember_me : BooleanField
        ...
    """

    email = StringField(
        "User",
        validators=[DataRequired()],
        filters=[strip_input, make_lower],
        render_kw=render.custom_field | {"placeholder": "E-mail address"},
        description="""
        Enter your E-mail address.
        """,
    )
    password = PasswordField(
        "Password",
        validators=[DataRequired()],
        render_kw=render.custom_field | {"placeholder": "Password"},
        description="""
        Enter your password.
        """,
    )
    remember_me = BooleanField(
        "Remember me",
        render_kw=render.boolean_field,
        description="""
        Do you want to be kept logged in until you actively log out or delete
        your browser cache?
        """,
    )
    submit = SubmitField("Sign In", render_kw=render.submit_field)



import re
from datetime import date
from itertools import chain, zip_longest

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqUtils import gc_fraction
from labbase2.models import BaseEntity, db
from labbase2.models.fields import SequenceString
from labbase2.models.mixins import Sequence
from sqlalchemy import asc, desc, func

__all__ = ["Oligonucleotide"]


class Oligonucleotide(BaseEntity, Sequence):
    """A class to represent a primer.

    Attributes
    ----------
    id : int
        The internal ID of this primer. This ID is unique among ALL entities in the
        database.
    date_ordered : date
        The date this oligo was ordered.
    sequence : str
        The sequence of the primer. Will be automatically converted to uppercase
        letters. Whitespaces are removed. Max length is 256.
    storage_place : str
        The storage location of the primer. Max length is 64.
    description : str
        A description of the purpose of the primer. Max length is 512.
    """

    __tablename__: str = "oligonucleotide"

    id = db.Column(
        db.Integer,
        db.ForeignKey("base_entity.id"),
        primary_key=True,
        info={"importable": False},
    )
    date_ordered = db.Column(db.Date, nullable=False, info={"importable": True})
    sequence = db.Column(SequenceString(256), nullable=False, info={"importable": True})
    storage_place = db.Column(db.String(64), info={"importable": True})
    description = db.Column(db.String(512), info={"importable": True})

    __mapper_args__ = {"polymorphic_identity": "oligonucleotide"}

    def __len__(self):
        return len(self.sequence)

    @property
    def gc_content(self) -> float:
        return gc_fraction(self.sequence)

    def formatted_seq(self, max_len: int = None) -> str:
        """Creates a formatted string for the sequence attribute by placing HTML
        <span> blocks around lowercase letters.

        Parameters
        ----------
        max_len : int
            The maximum number of bases that shall be returned. Default is 'None'
            which means all bases.

        Returns
        -------
        str
            The sequence of this primer with '<span class="lower-seq">'
            elements around blocks of lowercase letters.

        """

        seq = self.sequence[:max_len]
        if max_len and max_len < len(self):
            seq += "..."

        # Extract blocks of lowercase and uppercase letters from sequence.
        low = re.compile("[a-z]+").findall(seq)
        upp = re.compile("[A-Z.]+").findall(seq)

        low = [f'<span class="lower-seq">{s}</span>' for s in low]

        # Merge the sequences in correct order.
        ordered = (low, upp) if seq[0].islower() else (upp, low)

        return "".join(chain(*zip_longest(*ordered, fillvalue="")))

    @property
    def seqrecord(self) -> SeqRecord:
        seq = [L if L in "ATCG" else "." for L in self.sequence.upper()]

        return SeqRecord(
            Seq("".join(seq)),
            id=self.label,
            description=f"id={self.id};len={len(self)}",
        )

    @classmethod
    def _order_by(cls, order_by: str, ascending: bool) -> tuple:
        match order_by:
            case "length":
                field = func.length(cls.sequence)
            case _:
                field = getattr(cls, order_by)

        fnc = asc if ascending else desc

        return (fnc(field),)



from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm
from wtforms.fields import SelectField, StringField, TextAreaField
from wtforms.validators import DataRequired, Length

__all__ = ["EditDilution"]


class EditDilution(BaseForm):
    """Form to edit an antibody dilution.

    Attributes
    ----------
    application : SelectField
        The application this dilution was determined for,
        e.g. 'immunostaining' or 'western blot'.
    dilution : StringField
        The dilution itself. This should be something like 1:x. This is
        limited to 32 characters.
    reference : StringField
        A reference for this dilution. The number of characters is limited 512.
    """

    application = SelectField(
        label="Application",
        validators=[DataRequired(), Length(max=64)],
        choices=[
            ("immunostaining", "Immunostaining"),
            ("western blot", "Western blot"),
            ("immunoprecipitation", "Immunoprecipitation"),
        ],
        render_kw=render.select_field,
    )
    dilution = StringField(
        label="Dilution",
        validators=[DataRequired(), Length(max=32)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Dilution"},
    )
    reference = TextAreaField(
        label="Reference",
        validators=[DataRequired(), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field | {"rows": 8},
        description="Give a short description of the sample and condition you used.",
    )



from functools import partial

import pandas as pd
from flask import Blueprint, flash, redirect, render_template, request, url_for
from flask_login import current_user, login_required
from labbase2.models import (
    Antibody,
    BaseFile,
    Chemical,
    ColumnMapping,
    FlyStock,
    ImportJob,
    Oligonucleotide,
    Plasmid,
    db,
)
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.files.forms import UploadFile
from labbase2.views.files.routes import upload_file
from sqlalchemy.exc import IntegrityError

from .forms import MappingForm

__all__ = ["bp"]


bp = Blueprint("imports", __name__, url_prefix="/imports", template_folder="templates")


@bp.route("/", methods=["GET"])
@login_required
def index():
    return render_template(
        "imports/main.html", title="Pending imports", jobs=current_user.import_jobs
    )


@bp.route("/upload/<string:type_>", methods=["POST"])
@login_required
@permission_required("Upload files")
def upload(type_: str):
    form = UploadFile()

    match type_.lower():
        case "antibody":
            entity_cls = Antibody
        case "chemical":
            entity_cls = Chemical
        case "fly_stock":
            entity_cls = FlyStock
        case "oligonucleotide":
            entity_cls = Oligonucleotide
        case "plasmid":
            entity_cls = Plasmid
        case _:
            flash(f"Unknown type {type}", "danger")
            return redirect(request.referrer)

    if not form.validate():
        return redirect(request.referrer)

    file = upload_file(form, BaseFile)

    # Try to read the file.
    match file.path.suffix:
        case ".csv":
            read_fnc = pd.read_csv
        case ".xls" | ".xlsx":
            read_fnc = partial(pd.read_excel, engine="openpyxl")
        case _:
            flash("Unknown file type!", "danger")
            return redirect(request.referrer)

    try:
        read_fnc(file.path)
    except pd.errors.ParserError:
        flash("File could not be parsed properly!", "danger")
        db.session.delete(file)
        db.session.commit()
        return redirect(request.referrer)
    except pd.errors.EmptyDataError:
        flash("File is empty or has improper header!", "danger")
        db.session.delete(file)
        db.session.commit()
        return redirect(request.referrer)
    except Exception as error:
        flash(f"An unknown error occurred during file parsing! {error}", "danger")
        db.session.delete(file)
        db.session.commit()
        return redirect(request.referrer)

    # Now create the ImportJob.
    import_job = ImportJob(user_id=current_user.id, file_id=file.id, entity_type=type_)

    for field in entity_cls.importable_fields():
        column_mapping = ColumnMapping(mapped_field=field)
        import_job.mappings.append(column_mapping)

    db.session.add(import_job)
    db.session.commit()

    return redirect(url_for("imports.edit", id_=import_job.id))


@bp.route("/edit/<int:id_>", methods=["GET", "POST"])
@login_required
def edit(id_: int):
    import_job = ImportJob.query.get(id_)

    if import_job is None:
        flash(f"No import with ID {id_}!", "danger")
        return redirect(url_for(".index"))
    if import_job.user_id != current_user.id:
        flash(f"You are not authorized to work on this import!", "danger")
        return redirect(url_for(".index"))

    file = import_job.file

    match file.path.suffix:
        case ".csv":
            table = pd.read_csv(file.path)
        case ".xls" | ".xlsx":
            table = pd.read_excel(file.path, engine="openpyxl")
        case _:
            flash("Unknown file type!", "danger")
            return redirect(url_for(".index"))

    fields = [mapping.mapped_field for mapping in import_job.mappings]
    defaults = [mapping.input_column for mapping in import_job.mappings]
    choices = [(str(col), str(col)) for col in table.columns]
    form = MappingForm(data={"mapping": defaults}, fields=fields, choices=choices)

    if form.validate_on_submit():
        for mapping, field in zip(import_job.mappings, form.mapping):
            mapping.input_column = None if field.data == "None" else field.data
            db.session.commit()

    return render_template(
        "imports/edit_import.html",
        title="Import Oligonucleotides",
        job=import_job,
        table=table,
        form=form,
    )


@bp.route("/import/<int:id_>", methods=["GET", "POST"])
@login_required
def execute(id_: int):
    job = ImportJob.query.get(id_)

    if job.user_id != current_user.id:
        flash(Message.ERROR("You are not authorized to execute this import."))
        return redirect(url_for(".index"))

    match job.entity_type:
        case "antibody":
            entity_cls = Antibody
        case "chemical":
            entity_cls = Chemical
        case "fly_stock":
            entity_cls = FlyStock
        case "oligonucleotide":
            entity_cls = Oligonucleotide
        case "plasmid":
            entity_cls = Plasmid
        case _:
            flash(f"Unknown entity type {job.entity_type}!", "danger")
            return redirect(request.referrer)

    mappings = ColumnMapping.query.filter(
        ColumnMapping.job_id == job.id, ColumnMapping.input_column.isnot(None)
    )
    fields = [mapping.mapped_field for mapping in mappings]
    colmns = [mapping.input_column for mapping in mappings]

    file = job.file

    match file.path.suffix:
        case ".csv":
            read_fnc = pd.read_csv
        case ".xls" | ".xlsx":
            read_fnc = partial(pd.read_excel, engine="openpyxl")
        case _:
            flash("Unknown file type!", "danger")
            return redirect(request.referrer)

    table = read_fnc(file.path)

    table = table[colmns]
    table.columns = fields

    successfull = 0
    unsuccessfull = 0

    for row in table.itertuples(index=False):
        row = row._asdict()

        for key, value in row.items():
            if pd.isnull(value):
                row[key] = None
            if isinstance(value, str):
                row[key] = value.strip()

        origin = (
            f"Created from file {file.filename_exposed} by {current_user.username}."
        )
        entity = entity_cls(origin=origin, **row)

        try:
            db.session.add(entity)
            db.session.commit()
        except IntegrityError:
            unsuccessfull += 1
            db.session.rollback()
            continue
        except Exception as error:
            unsuccessfull += 1
            db.session.rollback()
            return redirect(url_for(".index"))
        else:
            successfull += 1

    db.session.delete(job)
    db.session.commit()

    return redirect(url_for(".index"))


@bp.route("/delete/<int:id_>", methods=["DELETE"])
@login_required
def delete(id_: int):
    if not (job := ImportJob.query.get(id_)):
        return Message.ERROR(f"No import with ID {id_}!")

    if job.user_id != current_user.id:
        return Message.ERROR("Only owner can delete import!")

    try:
        db.session.delete(job)
        db.session.commit()
    except Exception as err:
        db.session.rollback()
        return Message.ERROR(str(err))
    else:
        return Message.SUCCESS(f"Successfully deleted import {id_}!")



from labbase2.forms import BaseForm, FilterForm, render
from labbase2.forms.filters import strip_input
from labbase2.models import Chemical, User
from sqlalchemy import func
from wtforms.fields import (
    DateField,
    IntegerField,
    SelectField,
    StringField,
    TextAreaField,
)
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["FilterStockSolution", "EditStockSolution"]


class FilterStockSolution(FilterForm):
    id = IntegerField(
        "ID",
        validators=[Optional()],
        render_kw=render.custom_field | {"placeholder": "ID"},
    )
    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Label"},
        description="The label, i.e. the name of the entity.",
    )
    responsible_id = SelectField(
        "Responsible",
        choices=[(0, "All")],
        coerce=int,
        render_kw=render.select_field,
    )
    solvent = StringField(
        "Solvent",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "ddH2O"},
    )
    storage_place = StringField(
        "Storage place",
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Freezer..."},
    )
    order_by = SelectField(
        label="Order by",
        choices=[
            ("label", "Label"),
            ("id", "ID"),
            ("user_id", "User"),
            ("solvent", "Solvent"),
        ],
        default="label",
        render_kw=render.select_field,
        description="The column by which the results shall be ordered.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        user = (
            User.query.join(Chemical)
            .with_entities(User.id, User.username)
            .group_by(User.username)
            .having(func.count(User.id) > 0)
            .order_by(User.username)
        )
        self.responsible_id.choices += user

    def fields(self) -> list:
        return [
            self.id,
            self.label,
            self.responsible_id,
            self.solvent,
            self.storage_place,
            self.order_by,
            self.ascending,
        ]


class EditStockSolution(BaseForm):

    solvent = StringField(
        label="Solvent",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Solvent"},
    )
    concentration = StringField(
        label="Concentration",
        validators=[DataRequired(), Length(max=32)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Concentration"},
    )
    storage_place = StringField(
        label="Storage Location",
        validators=[DataRequired(), Length(max=64)],
        render_kw=render.custom_field | {"placeholder": "Storage location"},
    )
    date_created = DateField(
        "Created",
        validators=[DataRequired()],
        render_kw=render.custom_field | {"type": "date"},
    )
    date_emptied = DateField(
        "Emptied at",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    details = TextAreaField(
        label="Concentration",
        validators=[Optional(), Length(max=2048)],
        render_kw=render.custom_field
        | {"placeholder": "Further details...", "size": 8},
    )



from datetime import date

from flask import render_template
from flask_login import current_user
from labbase2.forms import EditEntityForm, render
from labbase2.forms.filters import remove_whitespaces, strip_input
from labbase2.forms.validators import AllowCharacters
from labbase2.models import User
from wtforms.fields import DateField, SelectField, StringField, TextAreaField
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["EditOligonucleotide"]


class EditOligonucleotide(EditEntityForm):
    date_ordered = DateField(
        "Order date",
        default=date.today,
        validators=[Optional()],
        render_kw=render.custom_field
        | {"type": "date", "placeholder": "Primer name, e.g. oRS-1"},
    )
    owner_id = SelectField(
        "Owner",
        validators=[DataRequired()],
        validate_choice=False,
        default=lambda: current_user.id,
        coerce=int,
        render_kw=render.select_field,
        description="""
        Be aware that you cannot edit this entry anymore if you select someone else.
        """,
    )
    sequence = StringField(
        "Sequence",
        validators=[DataRequired(), Length(max=256), AllowCharacters("ACGTacgt")],
        filters=[strip_input, remove_whitespaces],
        render_kw=render.custom_field | {"placeholder": "Sequence"},
    )
    storage_place = StringField(
        "Storage place",
        validators=[Optional(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Storage place"},
    )
    description = TextAreaField(
        "Description",
        validators=[Optional(), Length(max=512)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Description", "rows": 6},
        description="""
        Give a short description about the purpose of this oligonucleotide.
        """,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        choices = User.query.with_entities(User.id, User.username)
        choices = [tuple(choice) for choice in choices]
        self.owner_id.choices = choices

    def render(self, action: str = "", method: str = "GET") -> str:
        return render_template(
            "oligonucleotides/form.html", form=self, action=action, method=method
        )



from flask import Blueprint
from flask import current_app as app
from flask import flash, redirect, render_template, request
from flask_login import current_user, login_required
from labbase2.models import BaseFile, FlyStock, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.comments.forms import EditComment
from labbase2.views.files.forms import UploadFile
from labbase2.views.files.routes import upload_file
from labbase2.views.requests.forms import EditRequest

from .forms import EditFlyStock, FilterFlyStocks

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "flystocks", __name__, url_prefix="/fly-stocks", template_folder="templates"
)

# bp.register_blueprint(modifications.bp)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterFlyStocks(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = FlyStock.filter_(**data)
    except Exception as err:
        flash(str(err), "danger")
        entities = FlyStock.filter_(order_by="label")

    return render_template(
        "fly_stocks/main.html",
        filter_form=form,
        import_file_form=UploadFile(),
        add_form=EditFlyStock(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=FlyStock.query.count(),
        title="Fly Stocks",
    )


@bp.route("/", methods=["POST"])
@login_required
@permission_required("Add Fly Stock")
def add():
    form = EditFlyStock()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    fly_stock = FlyStock()
    form.populate_obj(fly_stock)

    try:
        db.session.add(fly_stock)
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully added stock '{fly_stock.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add Fly Stock")
def edit(id_: int):
    form = EditFlyStock()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (fly_stock := FlyStock.query.get(id_)) is None:
        return Message.ERROR(f"No fly stock with ID {id_}!")

    if fly_stock.owner_id != current_user.id:
        return Message.ERROR("Fly stocks can only be edited by owner!")

    form.populate_obj(fly_stock)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully edited fly stock!")


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add Fly Stock")
def delete(id_):
    if (fly_stock := FlyStock.query.get(id_)) is None:
        return Message.ERROR(f"No fly stock with ID {id_}!")

    if fly_stock.owner_id != current_user.id:
        return Message.ERROR("Only owner can delete fly stock!")

    try:
        db.session.delete(fly_stock)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully deleted fly stock '{fly_stock.label}'!")


@bp.route("/<int:id_>", methods=["GET"])
@login_required
def details(id_: int):
    if (fly_stock := FlyStock.query.get(id_)) is None:
        return Message.ERROR(f"No fly stock with ID {id_}!")

    return render_template(
        "fly_stocks/details.html",
        flystock=fly_stock,
        form=EditFlyStock(None, obj=fly_stock),
        file_form=UploadFile,
        comment_form=EditComment,
        request_form=EditRequest,
    )


@bp.route("/export/<string:format_>/", methods=["GET"])
@login_required
@permission_required("Export content")
def export(format_: str):
    data = FilterFlyStocks(request.args).data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = FlyStock.filter_(**data)
    except Exception as error:
        return Message.ERROR(error)

    match format_:
        case "csv":
            return FlyStock.export_to_csv(entities)
        case "json":
            return FlyStock.export_to_json(entities)
        case _:
            return Message.ERROR(f"Unsupported format '{format_}'!")



from app.forms.utils import err2message
from app.models import Modification, db
from flask import Blueprint
from flask_login import current_user, login_required

from .forms import EditModification

__all__: list[str] = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "modifications", __name__, url_prefix="/modifications", template_folder="templates"
)


@bp.route("/<int:flystock_id>", methods=["POST"])
@login_required
def add(flystock_id: int):
    if (form := EditModification()).validate():
        modification = Modification(fly_id=flystock_id, user_id=current_user.id)
        form.populate_obj(modification)

        try:
            db.session.add(modification)
            db.session.commit()
        except Exception as error:
            return str(error), 400
        else:
            return "Successfully added modification!", 201
    else:
        return err2message(form.errors), 400


@bp.route("/<int:id>", methods=["PUT"])
@login_required
def edit(id: int):
    if (form := EditModification()).validate():
        if (modification := Modification.query.get(id)) is None:
            return f"No preparation with ID {id}!", 404
        elif modification.user_id != current_user.id:
            return "Modification can only be edited by source user!", 400
        else:
            form.populate_obj(modification)

        try:
            db.session.commit()
        except Exception as error:
            return str(error), 400
        else:
            return f"Successfully edited modification!", 200
    else:
        return err2message(form.errors), 400


@bp.route("/<int:flystock_id>/<int:id>", methods=["DELETE"])
@login_required
def delete(flystock_id: int, id: int):
    if (modification := Modification.query.get(id)) is None:
        return f"No modification with ID {id}!", 404
    elif modification.user_id != current_user.id:
        return "Modification can only be deleted by source user!", 400
    else:
        try:
            db.session.delete(modification)
            db.session.commit()
        except Exception as error:
            return str(error), 400
        else:
            return f"Successfully deleted modification {id}!", 200



from app.forms.filters import strip_input
from app.forms.forms import EditForm
from app.forms.utils import RENDER_KW
from wtforms.fields import DateField, TextAreaField
from wtforms.validators import DataRequired, Length

__all__: list[str] = ["EditModification"]


class EditModification(EditForm):

    date = DateField(
        label="Date",
        validators=[DataRequired()],
        render_kw=RENDER_KW | {"id": "edit-form-modification-date", "type": "date"},
    )
    description = TextAreaField(
        label="Description",
        validators=[DataRequired(), Length(max=2048)],
        filters=[strip_input],
        render_kw=RENDER_KW
        | {
            "id": "edit-form-modification-description",
            "placeholder": "Description",
            "rows": 8,
        },
    )



from enum import Enum
from typing import Any


class Message(Enum):
    """
    An Enum for different message types.
    """

    INFO = "info"
    SUCCESS = "success"
    WARNING = "warning"
    ERROR = "danger"

    def __init__(self, cls: str):
        self.div = '<div class="alert alert-' + cls + '">{msg}</div>'

    def __call__(self, msg: Any) -> str:
        """Show message in a custom DIV block.

        Parameters
        ----------
        msg : Any
            The message that should be wrapped into an alert DIV. This must be a str or
            must be convertible to a str.

        Returns
        -------
        str
            A str that display the message in a custom DIV block.
        """

        return self.div.format(msg=msg)



import zoneinfo

from flask import current_app
from labbase2.forms import BaseForm, render
from wtforms.fields import (
    FileField,
    PasswordField,
    SelectField,
    StringField,
    SubmitField,
)
from wtforms.validators import DataRequired, Email, Length

__all__ = ["EditUserForm"]


class EditUserForm(BaseForm):
    """A registration form for new users.

    Attributes
    ----------
    username : StringField
    email : StringField
    roles : SelectField
    password : PasswordField
    password2 : PasswordField
    submit : SubmitField

    Notes
    -----
    The registration form is thought to be exposed only to an administrator
    of the site since the users should not be allowed to choose their own roles.
    """

    first_name = StringField(
        label="First name",
        validators=[DataRequired(), Length(max=64)],
        render_kw=render.custom_field | {"placeholder": "First name"},
        description="You given name. You may use initials for middle names.",
    )
    last_name = StringField(
        label="Last name",
        validators=[DataRequired(), Length(max=64)],
        render_kw=render.custom_field | {"placeholder": "First name"},
    )
    email = StringField(
        label="E-Mail Address",
        validators=[DataRequired(), Email(), Length(max=128)],
        render_kw=render.custom_field
        | {"id": "register-form-email", "placeholder": "Email Address"},
        description="The university email address.",
    )
    timezone = SelectField(
        "Timezone",
        choices=[(tz, tz) for tz in sorted(zoneinfo.available_timezones())],
        default=lambda: current_app.config["DEFAULT_TIMEZONE"],
        validators=[DataRequired()],
        render_kw=render.select_field,
        description="Select the timezone in which times shall be displayed for you.",
    )
    file = FileField(
        "Picture",
        render_kw=render.file_field,
        description="Ideally use a portrait with 1:1 aspect ratio.",
    )
    password = PasswordField(
        "Password",
        validators=[DataRequired()],
        render_kw=render.custom_field | {"placeholder": "Password"},
        description="Verify your password.",
    )



from flask import render_template
from flask_wtf import FlaskForm
from labbase2.forms import render
from labbase2.forms.filters import make_upper, strip_input
from labbase2.forms.validators import AllowCharacters
from wtforms.fields import BooleanField, IntegerField, SubmitField, TextAreaField
from wtforms.validators import DataRequired, Length, NumberRange

__all__ = ["FindOligonucleotide"]


class FindOligonucleotide(FlaskForm):
    sequence = TextAreaField(
        "Target sequence",
        validators=[DataRequired(), Length(max=20_000), AllowCharacters("ACTGactg")],
        filters=[strip_input, make_upper],
        render_kw=render.custom_field
        | {"size": 4, "Placeholder": "Target sequence..."},
        description="""
        The target sequence, for which a matching primer shall be found. Please note,
        that long sequence will considerably take longer to search. Therefor, 
        consider to restrict the search to a part of the sequence if possible.
        """,
    )
    min_match = IntegerField(
        "Minimum continuous match",
        validators=[DataRequired(), NumberRange(max=40)],
        default=20,
        render_kw=render.custom_field | {"size": 4, "placeholder": "Minimum match"},
        description="""
        The minimum CONTINUOUS match length such that the primer shall be considered
        'matching'. Lower numbers result in less specific primers but speed up the 
        search.
        """,
    )
    max_len = IntegerField(
        "Maximal primer length",
        validators=[DataRequired()],
        default=40,
        render_kw=render.custom_field | {"size": 4, "placeholder": "Start"},
        description="""
        Consider only primers that are at most this long. This excludes unlikely
        oligonucleotides (for instance for PCR). Lower numbers speed up the search.
        """,
    )
    reverse_complement = BooleanField(
        "Reverse complement",
        default=False,
        render_kw=render.boolean_field,
        description="""
        Specificy if the original sequence or the reverse complement shall be queried.
        If checked, the target sequence will be turned into the reverse complement 
        before searching.
        """,
    )
    submit = SubmitField("Search", render_kw=render.submit_field)

    def fields(self) -> list:
        return [self.sequence, self.min_match, self.max_len, self.reverse_complement]

    def render(self, action: str = "", method: str = "POST") -> str:
        return render_template(
            "forms/filter.html", form=self, action=action, method=method, csrf=True
        )



import json
import secrets

from email_validator import validate_email

__all__ = ["generate"]


def generate():
    secret_key = secrets.token_hex()

    while True:
        first_name = input("\nEnter the first name of an admin: ").strip()

        if 0 < len(first_name) < 64:
            break
        else:
            print("The username is too short or too long.")

    while True:
        last_name = input("\nEnter the last name of an admin: ").strip()

        if 0 < len(last_name) < 64:
            break
        else:
            print("The username is too short or too long.")

    while True:
        email = input("\nEnter the email address of an admin: ").strip()

        if not validate_email(email):
            print("The email address is invalid!")
        else:
            break

    for_json = {"SECRET_KEY": secret_key, "USER": (first_name, last_name, email)}

    with open("config.json", "w") as file:
        file.write(json.dumps(for_json))



from flask_wtf import FlaskForm
from labbase2.forms import render
from wtforms.fields import BooleanField, SubmitField

__all__ = ["EditPermissions"]


class EditPermissions(FlaskForm):

    write_comment = BooleanField(
        "Write Comments", default=False, render_kw=render.boolean_field
    )
    upload_files = BooleanField(
        "Upload Files", default=False, render_kw=render.boolean_field
    )
    add_dilutions = BooleanField(
        "Add Dilutions", default=False, render_kw=render.boolean_field
    )
    add_preparations = BooleanField(
        "Add Preparations", default=False, render_kw=render.boolean_field
    )
    add_glycerol_stocks = BooleanField(
        "Add Glycerol Stocks", default=False, render_kw=render.boolean_field
    )
    add_consumable_batches = BooleanField(
        "Add Batches", default=False, render_kw=render.boolean_field
    )
    add_antibodies = BooleanField(
        "Add Antibodies", default=False, render_kw=render.boolean_field
    )
    add_plasmid = BooleanField(
        "Add plasmids", default=False, render_kw=render.boolean_field
    )
    add_oligonucleotide = BooleanField(
        "Add Oligonucleotides", default=False, render_kw=render.boolean_field
    )
    manage_users = BooleanField(
        "Manage Users", default=False, render_kw=render.boolean_field
    )
    export_content = BooleanField(
        "Export content", default=False, render_kw=render.boolean_field
    )
    add_requests = BooleanField(
        "Add requests", default=False, render_kw=render.boolean_field
    )

    submit = SubmitField("Update permissions", render_kw=render.submit_field)
